/*
     This file is part of GNUnet.
     (C) 2001, 2002 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * @file applications/afs/protocol/handler.c
 * @author Christian Grothoff
 *
 * Handlers for incoming AFS requests (p2p and CS).
 **/

#include "config.h"
#include "util/tcpio.h"
#include "util/tcp_return.h"
#include "util/statistics.h"
#include "applications/afs/database/fileindex.h"
#include "applications/afs/database/manager.h"
#include "applications/afs/protocol/messages.h"
#include "applications/afs/protocol/handler.h"
#include "applications/afs/protocol/policy.h"
#include "applications/afs/protocol/afs.h"
#include <stdio.h>

/* ********************* p2p handlers ****************** */

static int stat_p2p_query_count;
static int stat_p2p_chk_replies;
static int stat_p2p_3hash_replies;
static int stat_cs_query_count;
static int stat_cs_insert_chk_count;
static int stat_cs_insert_3hash_count;
static int stat_cs_index_block_count;
static int stat_cs_index_file_count;
static int stat_cs_index_super_count;

/**
 * Is it ok to use the bloomfilter to speed-up
 * insertion of content that is already present,
 * or can the bloomfilter not be trusted? (e.g.
 * gnunet-check-mode).
 **/
static int fastInsert;

/**
 * Initialize the handler module. Registers counters
 * with the statistics module.
 *
 * @return OK on success, SYSERR on failure
 **/
int initAFSHandler() {
  fastInsert = testConfigurationString("AFS",
				       "FAST-INSERT",
				       "YES");
  stat_p2p_query_count 
    = statHandle("# p2p queries received");
  stat_p2p_chk_replies 
    = statHandle("# p2p CHK content received (kb)");
  stat_p2p_3hash_replies 
    = statHandle("# p2p search results received (kb)");
  stat_cs_query_count 
    = statHandle("# client queries received");
  stat_cs_insert_chk_count 
    = statHandle("# client CHK content inserted (kb)");
  stat_cs_insert_3hash_count 
    = statHandle("# client 3HASH search results inserted (kb)");
  stat_cs_index_block_count 
    = statHandle("# client file index requests received");
  stat_cs_index_file_count 
    = statHandle("# file index requests received");
  stat_cs_index_super_count 
    = statHandle("# super query index requests received");
  return OK;
}

/**
 * Handle query for content. Depending on how we like the sender,
 * lookup, forward or even indirect.
 **/
int handleQUERY(HostIdentity * sender,
		p2p_HEADER * msg) {
  QUERY_POLICY qp;
  AFS_p2p_QUERY * qmsg;
  int queries;
  
  queries = (ntohs(msg->size) - sizeof(AFS_p2p_QUERY)) / sizeof(HashCode160);
  if ( (queries <= 0) || 
       (ntohs(msg->size) != sizeof(AFS_p2p_QUERY) + queries * sizeof(HashCode160)) ) {
    LOG(LOG_WARNING,
	"WARNING: query received was malformed\n");
    return SYSERR;
  }
  statChange(stat_p2p_query_count, 1);
  qmsg = (AFS_p2p_QUERY*) msg;
  qmsg->ttl 
    = htonl(ntohl(qmsg->ttl) - TTL_DECREMENT);
  if (((signed int)ntohl(qmsg->ttl)) <= 0)
    return OK;  

  qp = evaluateQuery(sender,
		     ntohl(qmsg->priority));  
  if ((qp & QUERY_DROPMASK) == 0)
    return OK; /* straight drop. */

  execQuery(qp, qmsg, NULL);
  return OK;
}
 
/**
 * Receive content, do something with it!  There are 3 basic
 * possiblilities. Either our node did the request and we should send
 * the result to a client via TCP, or the content was requested by
 * another node and we forwarded the request (and thus we now have to
 * fwd the reply) or 3rd somebody just send us some content we did NOT
 * ask for - and we can choose to store it or just discard it.
 **/
int handleCHK_CONTENT(HostIdentity * sender, 
		      p2p_HEADER * msg) {
  int prio;
  HashCode160 queryHash;
  ContentIndex ce;
  AFS_p2p_CHK_RESULT * cmsg;
  int ret;
  int dupe;

  if (ntohs(msg->size) != sizeof(AFS_p2p_CHK_RESULT)) {
    LOG(LOG_WARNING,
	"WARNING: CHK content message received was malformed\n");
    return SYSERR;
  }
  statChange(stat_p2p_chk_replies, 1);
  cmsg = (AFS_p2p_CHK_RESULT*) msg;
  hash(&cmsg->result,
       CONTENT_SIZE,
       &queryHash);
  prio = useContent(sender,
		    &queryHash,
		    msg);
  if (sender == NULL) /* no migration, this is already content
			 from the local node */
    return OK;  
  prio = evaluateContent(&queryHash,
			 prio);
  if (prio == SYSERR)
    return OK; /* straight drop */
  memcpy(&ce.hash,
	 &queryHash,
	 sizeof(HashCode160));
  ce.importance    = htonl(prio);
  ce.type          = htons(LOOKUP_TYPE_CHK);
  ce.fileNameIndex = 0;
  ce.fileOffset    = 0;
  ret = insertContent(&ce, 
		      sizeof(CONTENT_Block),
		      &cmsg->result, 
		      sender,
		      &dupe);
  if (ret == OK)
    if (dupe == NO || NO == testBloomfilter(singleBloomFilter,
     			  	  	    &queryHash))
      addToBloomfilter(singleBloomFilter,
	  	       &queryHash);
  return OK;
}

/**
 * Receive content, do something with it!  There are 3 basic
 * possiblilities. Either our node did the request and we should send
 * the result to a client via TCP, or the content was requested by
 * another node and we forwarded the request (and thus we now have to
 * fwd the reply) or 3rd somebody just send us some content we did NOT
 * ask for - and we can choose to store it or just discard it.
 **/
int handle3HASH_CONTENT(HostIdentity * sender, 
			p2p_HEADER * msg) {
  int prio;
  AFS_p2p_3HASH_RESULT * cmsg;
  HashCode160 tripleHash;
  ContentIndex ce;
  HexName hex;
  int ret;
  int dupe;

  if (ntohs(msg->size) != sizeof(AFS_p2p_3HASH_RESULT)) {
    LOG(LOG_WARNING,
	"WARNING: content message received was malformed\n");
    return SYSERR;
  }
  statChange(stat_p2p_3hash_replies, 1);
  cmsg = (AFS_p2p_3HASH_RESULT*) msg;
  hash(&cmsg->hash,
       sizeof(HashCode160),
       &tripleHash);
  IFLOG(LOG_DEBUG,
	hash2hex(&tripleHash,
		 &hex));
  LOG(LOG_DEBUG,
      "DEBUG: received 3HASH search result for %s from peer\n",
      &hex);
  prio = useContent(sender,
		    &tripleHash,
		    msg);
  if (sender == NULL) { /* no migration, this is already content
			   from the local node */
    LOG(LOG_DEBUG,
	"DEBUG: content migration not needed, content is local\n");
    return OK;  
  } else
    LOG(LOG_DEBUG,
	"DEBUG: content migration with preference %d\n",
	prio);
  prio = evaluateContent(&tripleHash,
			 prio);
  if (prio == SYSERR) {
    LOG(LOG_DEBUG,
	"DEBUG: content not important enough, not replicated\n");
    return OK; /* straight drop */
  } else
    LOG(LOG_DEBUG,
	"DEBUG: content replicated with total preference %d\n",
	prio);
  memcpy(&ce.hash,
	 &cmsg->hash,
	 sizeof(HashCode160));
  ce.importance    = htonl(prio);
  ce.type          = htons(LOOKUP_TYPE_3HASH);
  ce.fileNameIndex = 0;
  ce.fileOffset    = 0;
  ret = insertContent(&ce, 
		      sizeof(CONTENT_Block),
             	      &cmsg->result, 
                      sender,
		      &dupe);
  if (ret == OK)
    if (dupe == NO || NO == testBloomfilter(singleBloomFilter,
     		  	    	  	    &tripleHash))
      addToBloomfilter(singleBloomFilter,
	               &tripleHash);
  return OK;
}

/* *********************** CS handlers ***************** */

/**
 * Process a query from the client. Forwards to the network.
 * @return SYSERR if the TCP connection should be closed, otherwise OK
 **/ 
int csHandleRequestQuery(GNUNET_TCP_SOCKET * sock,
			 AFS_CS_QUERY * queryRequest) {
  QUERY_POLICY qp = QUERY_ANSWER|QUERY_FORWARD|QUERY_INDIRECT|QUERY_PRIORITY_BITMASK; 
  AFS_p2p_QUERY * msg;
  HexName hex;
  int queries;

  queries = (ntohs(queryRequest->header.size) - sizeof(AFS_CS_QUERY)) / sizeof(HashCode160);
  if ( (queries <= 0) ||
       (ntohs(queryRequest->header.size) != sizeof(AFS_CS_QUERY) + queries * sizeof(HashCode160)) ) {
    LOG(LOG_WARNING,
	"WARNING: received malformed query from client\n");
    return SYSERR;
  }
  statChange(stat_cs_query_count, 1);
  IFLOG(LOG_DEBUG,
	hash2hex(&queryRequest->queries[0], 
		 &hex));
  LOG(LOG_DEBUG, 
      "DEBUG: received %d queries (%s) with ttl %d and priority %d.\n",
      queries,
      &hex,
      ntohl(queryRequest->ttl),
      ntohl(queryRequest->priority));
  msg = xmalloc(sizeof(AFS_p2p_QUERY)+queries * sizeof(HashCode160),
		"allocating space for p2p query message");
  msg->header.size 
    = htons(sizeof(AFS_p2p_QUERY)+queries * sizeof(HashCode160));
  msg->header.requestType 
    = htons(AFS_p2p_PROTO_QUERY);
  memcpy(&msg->queries[0],
	 &queryRequest->queries[0],
	 sizeof(HashCode160) * queries);
  msg->priority 
    = queryRequest->priority; /* no htonl here: is already in network byte order! */
  msg->ttl 
    = queryRequest->ttl; /* no htonl here: is already in network byte order! */
  memcpy(&msg->returnTo,
	 coreAPI->myIdentity,
	 sizeof(HostIdentity));
  execQuery(qp, msg, sock);   
  xfree(msg,
	"freeing query message");
  return OK;
}

/**
 * Process a request to insert content from the client.
 * @return SYSERR if the TCP connection should be closed, otherwise OK
 **/
int csHandleRequestInsertCHK(GNUNET_TCP_SOCKET * sock,
			     AFS_CS_INSERT_CHK * insertRequest) {
  ContentIndex entry;
  HexName hex;
  int ret;
  int dupe;

  if (ntohs(insertRequest->header.size) != 
      sizeof(AFS_CS_INSERT_CHK)) {
    LOG(LOG_WARNING,
	"WARNING: received malformed CHK insert request from client\n");
    return SYSERR;
  } 
  statChange(stat_cs_insert_chk_count, 1);
   hash(&insertRequest->content,
       sizeof(CONTENT_Block),
       &entry.hash);
  IFLOG(LOG_DEBUG,
	hash2hex(&entry.hash,
		 &hex));
  LOG(LOG_DEBUG,
      "DEBUG: received CHK insert request for block %s\n",
      &hex);
  entry.type
    = htons(LOOKUP_TYPE_CHK);
  entry.importance
    = insertRequest->importance; /* both are in network byte order! */
  entry.fileNameIndex 
    = 0; /* database */
  entry.fileOffset 
    = 0; /* data/content */

  if ( (NO == fastInsert) ||
       (NO == testBloomfilter(singleBloomFilter,
			      &entry.hash)) ) {
    ret = insertContent(&entry,
			sizeof(CONTENT_Block),
			&insertRequest->content,
			NULL,
		        &dupe);
    if (ret == OK)
      if (dupe == NO || NO == testBloomfilter(singleBloomFilter,
     	  	     	       	  	      &entry.hash))
        addToBloomfilter(singleBloomFilter,
		         &entry.hash);
  } else
    ret = OK;
  return sendTCPResult(sock, ret);
}

/**
 * Process a request to insert content from the client.
 * @return SYSERR if the TCP connection should be closed, otherwise OK
 **/
int csHandleRequestInsert3HASH(GNUNET_TCP_SOCKET * sock,
			       AFS_CS_INSERT_3HASH * insertRequest) {
  ContentIndex entry;
  HashCode160 tripleHash;
  HexName hex;
  int dupe;
  int ret;

  if (ntohs(insertRequest->header.size) != 
      sizeof(AFS_CS_INSERT_3HASH)) {
    LOG(LOG_WARNING,
	"WARNING: received malformed 3HASH insert request from client\n");
    return SYSERR;
  }
  statChange(stat_cs_insert_3hash_count, 1);
  memcpy(&entry.hash,
	 &insertRequest->doubleHash,
	 sizeof(HashCode160));
  hash(&insertRequest->doubleHash,
       sizeof(HashCode160),
       &tripleHash);
  IFLOG(LOG_DEBUG,
	hash2hex(&tripleHash,
		 &hex));
  LOG(LOG_DEBUG,
      "DEBUG: received 3HASH insert request for %s from client\n",
      &hex);
  entry.type
    = htons(LOOKUP_TYPE_3HASH);
  entry.importance
    = insertRequest->importance; /* both are in network byte order! */
  entry.fileNameIndex 
    = 0; /* database */
  entry.fileOffset 
    = 0; /* data/content */
  ret = insertContent(&entry,
		      sizeof(CONTENT_Block),
		      &insertRequest->content,
	   	      NULL,
		      &dupe);
  if (ret == OK)     
    if (dupe == NO || NO == testBloomfilter(singleBloomFilter,
     	              	       	  	    &tripleHash))
      addToBloomfilter(singleBloomFilter,
		       &tripleHash);
		     
  return sendTCPResult(sock, ret);
}

/**
 * Process a request to index content from the client.
 * @return SYSERR if the TCP connection should be closed, otherwise OK
 **/
int csHandleRequestIndexBlock(GNUNET_TCP_SOCKET * sock,
			      AFS_CS_INDEX_BLOCK * indexingRequest) {
  if (ntohs(indexingRequest->header.size) != 
      sizeof(AFS_CS_INDEX_BLOCK)) {
    LOG(LOG_WARNING, 
	"WARNING: block indexing request from client was malformed!\n");
    return SYSERR;
  }
  statChange(stat_cs_index_block_count, 1);
  return sendTCPResult(sock,
		       addEntry(&indexingRequest->contentIndex,
				NULL));
}

/**
 * Process a query to list a file as on-demand encoded from the client.
 *
 * @return SYSERR if the TCP connection should be closed, otherwise OK
 **/
int csHandleRequestIndexFile(GNUNET_TCP_SOCKET * sock,
			     AFS_CS_INDEX_FILE * listFileRequest) {
  if (ntohs(listFileRequest->header.size) != 
      sizeof(AFS_CS_INDEX_FILE)) {
    LOG(LOG_WARNING, 
	"WARNING: file indexing request from client was malformed!\n");
    return SYSERR;
  }
  statChange(stat_cs_index_file_count, 1);
  listFileRequest->filename[1023] = 0; /* ensure strlen terminates! */
  return sendTCPResult(sock, 
		       appendFilename(listFileRequest->filename));
}

/**
 * Process a client request to extend our super-query bloom
 * filter.
 **/
int csHandleRequestIndexSuper(GNUNET_TCP_SOCKET * sock,
			      AFS_CS_INDEX_SUPER * superIndexRequest) {
  ContentIndex entry;

  if (ntohs(superIndexRequest->header.size) != 
      sizeof(AFS_CS_INDEX_SUPER)) {
    LOG(LOG_WARNING, 
	"WARNING: super-hash indexing request from client was malformed!\n");
    return SYSERR;
  }
  statChange(stat_cs_index_super_count, 1);
  if (ntohl(superIndexRequest->isSuper) == YES) {
    addToBloomfilter(superBloomFilter,
		     &superIndexRequest->superHash);
    entry.type
      = htons(LOOKUP_TYPE_SUPER);
    entry.importance
      = 0; /* super-queries are always incredbily important */
    entry.fileNameIndex 
      = 0; /* database */
    entry.fileOffset 
      = 0; /* data/content */
    memcpy(&entry.hash,
	   &superIndexRequest->superHash,
	   sizeof(HashCode160));
    return sendTCPResult(sock, 
			 addEntry(&entry,
				  NULL));
  } else {
    addToBloomfilter(singleBloomFilter,
		     &superIndexRequest->superHash);
    return sendTCPResult(sock, OK);
  }
}


/* end of handler.c */
