/*
     This file is part of GNUnet.
     (C) 2001, 2002 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * @file applications/afs/tools/gnunet-convert.c
 * @author Igor Wronsky
 *
 * Little tool to convert content databases from format to another.
 * Use if you change the database manager type or the bucket count.
 *
 **/

#include "config.h"

#include <sys/types.h>
#include <unistd.h>
#include <pwd.h>
#include <pthread.h>
#include <signal.h>
#include <unistd.h>
#include <errno.h>
#include "util/getopt.h"
#include "util/tcp_return.h"
#include "util/tcpio.h"
#include "util/statistics.h"
#include "util/ports.h"
#include "util/hashing.h"
#include "util/initialize.h"
#include "util/logging.h"
#include "applications/afs/afs.h"
#include "applications/afs/util/contentencoding.h"
#include "applications/afs/encoding/insertutil.h"
#include "applications/afs/database/lookup.h"
#include "applications/afs/database/fileindex.h"
#include "applications/afs/database/contentdatabase.h"
#include "applications/common/port.h"
#include "applications/common/printhelp.h"
  
static DatabaseHandle * srcHandle;
static DatabaseHandle * dstHandle;

unsigned int insertedBlocks = 0;
unsigned int failedBlocks = 0;
unsigned int otherBlocks = 0;

/**
 * To make the linker happy... (database library links against these symbols, 
 * typically defined in the AFS library, but we don't need to drag that
 * library in).
 **/
void * coreAPI;
DatabaseHandle * dbHandle;

static int progressDot = 0;
static int be_verbose = NO;
static int be_quiet = NO;

static int addToDestination(ContentIndex * entry,
			CollisionLevel * level,
			size_t pos,
			void * data) {
  progressDot++;
  if ((progressDot & 255) == 0)
    if (be_quiet == NO) {
      printf(".");
      fflush(stdout);
    }
  if (ntohs(entry->type) == LOOKUP_TYPE_SUPER) {
    otherBlocks++;
    return OK; /* no real content that corresponds to super-hashes */
  }

  if (ntohs(entry->fileNameIndex) == 0) {
    int res;
    void * block;

    res = readContent(srcHandle,
                      &entry->hash,
                      &block);
    if (res != SYSERR) {
      writeContent(dstHandle,
		   &entry->hash,
		   res,
		   block);
      xfree(block,
      	   "freeing block");
      insertedBlocks++;
    } else {
      failedBlocks++;
    }
  } else {
    otherBlocks++;
  }
  return OK;
}

/**
 * Print a list of the options we offer.
 **/
static void printhelp() {
  printf("USAGE: gnunet-convert [OPTIONS]\n"
	 "-S TYPE, --source=TYPE        : "
	 "type of source contentdatabase\n"
	 "-D TYPE, --dest=TYPE          : "
	 "type of destination contentdatabase\n"
	 "-b BUCKETS, --sbuckets=BUCKETS: "
	 "number of buckets in source db\n"
	 "-p BUCKETS, --dbuckets=BUCKETS : "
	 "number of buckets in destination db\n"
	 "-V, --verbose                 : "
	 "be verbose\n"
	 "-q, --quiet                   : "
	 "be quiet\n");
  printHelpDefault();
  printf("Never run gnunet-convert while gnunetd is running!\n");
}

/**
 * Perform option parsing from the command line. 
 **/
static int parseCommandLine(int argc, 
			    char * argv[]) {
  int shouldexit=0;
  int c;
  unsigned int tmp;

  freeNonNull(setConfigurationString("GNUNETD",
				     "LOGFILE",
				     NULL),
	      "gnunet-check always use option d");
  while (1) {
    int option_index = 0;
    static struct option long_options[] = {
      LONG_DEFAULT_OPTIONS,
      { "source",   1, 0, 'S' },
      { "sbuckets", 1, 0, 'b' },
      { "dest",     1, 0, 'D' },
      { "dbuckets", 1, 0, 'p' },
      { "verbose",  0, 0, 'V' },
      { "quiet",    0, 0, 'q' },
      { 0,0,0,0 }
    };
    
    c = getopt_long(argc,
		    argv, 
		    "vhdc:nS:b:D:p:VqL:", 
		    long_options, 
		    &option_index);    
    if (c == -1) 
      break;  // No more flags to process
    if (YES == parseDefaultOptions(c, optarg))
      continue;    
    switch(c) {
    case 'S':
      freeNonNull(setConfigurationString("GNUNET-CONVERT",
					 "SOURCETYPE",
					 optarg),
		  "gnunet-check option parsing S");
     break;
    case 'D':
      freeNonNull(setConfigurationString("GNUNET-CONVERT",
					 "DESTTYPE",
					 optarg),
		  "gnunet-check option parsing D");
     break;
    case 'b':
      if (1 != sscanf(optarg, "%ud", &tmp)) {
	LOG(LOG_FAILURE,
	    "FAILURE: You must pass a number to the -b option.\n");
	return -1;
      }
      
     setConfigurationInt("GNUNET-CONVERT",
			 "SOURCEBUCKETS",
			 tmp);
     break;
    case 'p':
      if (1 != sscanf(optarg, "%ud", &tmp)) {
	LOG(LOG_FAILURE,
	    "FAILURE: You must pass a number to the -p option.\n");
	return -1;
      }
      setConfigurationInt("GNUNET-CONVERT",
 	 	 	  "DESTBUCKETS",
 			  tmp);
     break;
    case 'q':
      be_quiet = YES;
      break;
    case 'V':
      be_verbose = YES;
      break;
    case 'v': 
      printf("GNUnet v%s, gnunet-convert v%s\n",
	     VERSION, AFS_VERSION);
      shouldexit=1;
      break;
    case 'h': 
      printhelp(); 
      shouldexit=1;
      break;
    default:
      printf("Unknown option %c. Aborting.\nUse --help to get a list of options.\n",
	    c);
      shouldexit=1;    
    } /* end of parsing commandline */
  }
  if (optind < argc) {
    printf("Invalid arguments: ");
    while (optind < argc)
      printf("%s ", argv[optind++]);
    printf("\nExiting.\n");
    return 1;
  }
  readConfiguration();
  return shouldexit;
}

int main(int argc, char * argv[]) {
  char * srcDb;
  char * dstDb;
  int srcBuckets;
  int dstBuckets;
  int entries;

  if (SYSERR == initUtil(argc, argv, &parseCommandLine))
    return 0;

  srcDb = getConfigurationString("GNUNET-CONVERT", 
				 "SOURCETYPE");
  dstDb = getConfigurationString("GNUNET-CONVERT", 
				 "DESTTYPE");
  srcBuckets = getConfigurationInt("GNUNET-CONVERT",
  				   "SOURCEBUCKETS");
  dstBuckets = getConfigurationInt("GNUNET-CONVERT",
  				   "DESTBUCKETS");
  if(srcDb == NULL || dstDb == NULL || srcBuckets < 1 || dstBuckets < 1) {
    fprintf(stderr,
	    "Invalid args.\n");
    return -1;
  }
  if(strcmp(srcDb, dstDb)==0) {
    fprintf(stderr,
    	    "Values of -D and -S can not be the same. See man page of gnunet-convert for getting around this.\n");
    return -1;
  }

  initLookup();

  srcHandle = initContentDatabase(srcDb, srcBuckets);
  dstHandle = initContentDatabase(dstDb, dstBuckets);
 
  entries = forEachLookupEntry(&addToDestination, NULL);

  fprintf(stdout, "\n==> Done processing %d entries in index "
  	          "(%d converted, %d missing, %d other).\n",
   	  	  entries,
		  insertedBlocks, failedBlocks, otherBlocks);

  doneContentDatabase(srcHandle);
  doneContentDatabase(dstHandle);

  doneLookup();
  doneUtil();

  return 0;
}


/* end of gnunet-convert.c */
